@@ -150,6 +150,7 @@ struct assignment_request {
 	uint16_t msc_rtp_port;
 	bool use_osmux;
 	uint8_t osmux_cid;
+	uint8_t rtp_extensions;
 
 	/* Rate/codec setting in preference order (need at least 1 !) */
 	int n_ch_mode_rate;
@@ -359,6 +360,7 @@ struct gsm_subscriber_connection {
 		/* RTP address where the MSC expects us to send the RTP stream coming from the BTS. */
 		char msc_assigned_rtp_addr[INET6_ADDRSTRLEN];
 		uint16_t msc_assigned_rtp_port;
+		uint8_t rtp_extensions;
 
 		/* The endpoint at the MGW used to join both BTS and MSC side connections, e.g.
 		 * "rtpbridge/23@mgw". */
@@ -526,9 +528,9 @@ struct gsm_abis_mo {
 	struct abis_om_obj_inst obj_inst;
 	const char *name;
 	struct gsm_nm_state nm_state;
-	struct tlv_parsed *nm_attr;
 	struct gsm_bts *bts;
 	struct osmo_fsm_inst *fi;
+	bool sw_act_rep_received;
 	bool opstart_sent;
 	bool adm_unlock_sent;
 	bool get_attr_sent;
@@ -538,6 +540,12 @@ struct gsm_abis_mo {
 	bool rsl_connect_sent;
 	bool rsl_connect_ack_received;
 	bool force_rf_lock;
+	/* vendor specific fields below */
+	union {
+		struct {
+			uint8_t obj_version;
+		} ipaccess;
+	};
 };
 
 /* Ericsson OM2000 Managed Object */
@@ -668,7 +676,6 @@ struct gsm_bts_trx_ts {
 	bool is_rsl_ready;
 
 	struct gsm_abis_mo mo;
-	struct tlv_parsed nm_attr;
 	uint8_t nm_chan_comb;
 	int tsc;		/* -1 == use BTS TSC */
 
@@ -892,6 +899,8 @@ void gsm_abis_mo_reset(struct gsm_abis_mo *mo);
 void gsm_mo_init(struct gsm_abis_mo *mo, struct gsm_bts *bts,
 		 uint8_t obj_class, uint8_t p1, uint8_t p2, uint8_t p3);
 
+struct gsm_abis_mo *gsm_objclass2mo(struct gsm_bts *bts, uint8_t obj_class,
+				    const struct abis_om_obj_inst *obj_inst);
 struct gsm_nm_state *
 gsm_objclass2nmstate(struct gsm_bts *bts, uint8_t obj_class,
 		 const struct abis_om_obj_inst *obj_inst);
@@ -1061,6 +1070,7 @@ struct gsm_network {
 
 	/* PCU socket state */
 	char *pcu_sock_path;
+	unsigned int pcu_sock_wqueue_len_max;
 	struct pcu_sock_state *pcu_state;
 };
 
