@@ -52,11 +52,15 @@
 #include <osmocom/bsc/nm_common_fsm.h>
 #include <osmocom/gsm/bts_features.h>
 #include <osmocom/bsc/ipaccess.h>
+#include <osmocom/bsc/bts_ipaccess_nanobts_omlattr.h>
 
 #define OM_ALLOC_SIZE		1024
 #define OM_HEADROOM_SIZE	128
 #define IPACC_SEGMENT_SIZE	245
 
+/* max number of SW Description IEs we can parse */
+#define SW_DESCR_MAX		5
+
 #define LOGPMO(mo, ss, lvl, fmt, args ...) \
 	LOGP(ss, lvl, "OC=%s(%02x) INST=(%02x,%02x,%02x): " fmt, \
 	     get_value_string(abis_nm_obj_class_names, (mo)->obj_class), \
@@ -560,6 +564,45 @@ static inline const uint8_t *parse_attr_resp_info_unreported(const struct abis_o
 	return ari + num_unreported + 1; /* we have to account for 1st byte with number of unreported attributes */
 }
 
+/* Parse Attribute Response Info content for 3GPP TS 52.021 §9.4.30 Manufacturer Id */
+static void parse_osmo_bts_features(struct gsm_bts *bts,
+				    const uint8_t *data, uint16_t data_len)
+{
+	/* log potential BTS feature vector overflow */
+	if (data_len > sizeof(bts->_features_data)) {
+		LOGPMO(&bts->mo, DNM, LOGL_NOTICE,
+		       "Get Attributes Response: feature vector is truncated "
+		       "(from %u to %zu bytes)\n", data_len, sizeof(bts->_features_data));
+		data_len = sizeof(bts->_features_data);
+	}
+
+	/* check that max. expected BTS attribute is above given feature vector length */
+	if (data_len > OSMO_BYTES_FOR_BITS(_NUM_BTS_FEAT)) {
+		LOGPMO(&bts->mo, DNM, LOGL_NOTICE,
+		       "Get Attributes Response: reported unexpectedly long (%u bytes) "
+		       "feature vector - most likely it was compiled against newer BSC headers. "
+		       "Consider upgrading your BSC to later version.\n", data_len);
+	}
+
+	memcpy(bts->_features_data, data, data_len);
+	bts->features_known = true;
+
+	/* Log each BTS feature in the reported vector */
+	for (unsigned int i = 0; i < data_len * 8; i++) {
+		if (!osmo_bts_has_feature(&bts->features, i))
+			continue;
+
+		if (i >= _NUM_BTS_FEAT) {
+			LOGPMO(&bts->mo, DNM, LOGL_NOTICE,
+			       "Get Attributes Response: unknown feature 0x%02x is supported\n", i);
+		} else {
+			LOGPMO(&bts->mo, DNM, LOGL_NOTICE,
+			       "Get Attributes Response: feature '%s' is supported\n",
+			       osmo_bts_features_name(i));
+		}
+	}
+}
+
 /* Handle 3GPP TS 52.021 §8.11.3 Get Attribute Response (with nanoBTS specific attribute formatting) */
 static int parse_attr_resp_info_attr(struct gsm_bts *bts, const struct gsm_bts_trx *trx, struct abis_om_fom_hdr *foh, struct tlv_parsed *tp)
 {
@@ -570,42 +613,22 @@ static int parse_attr_resp_info_attr(struct gsm_bts *bts, const struct gsm_bts_t
 	uint16_t port;
 	struct in_addr ia = {0};
 	char unit_id[40];
-	struct abis_nm_sw_desc sw_descr[MAX_BTS_ATTR];
 
-	/* Parse Attribute Response Info content for 3GPP TS 52.021 §9.4.30 Manufacturer Id */
-	if (bts->type == GSM_BTS_TYPE_OSMOBTS && TLVP_PRES_LEN(tp, NM_ATT_MANUF_ID, 2)) {
-		len = TLVP_LEN(tp, NM_ATT_MANUF_ID);
-
-		/* log potential BTS feature vector overflow */
-		if (len > sizeof(bts->_features_data)) {
-			LOGPMO(&bts->mo, DNM, LOGL_NOTICE, "Get Attributes Response: feature vector is truncated "
-			       "(from %u to %zu bytes)\n", len, sizeof(bts->_features_data));
-			len = sizeof(bts->_features_data);
+	switch (bts->type) {
+	case GSM_BTS_TYPE_OSMOBTS:
+		if (TLVP_PRES_LEN(tp, NM_ATT_MANUF_ID, 2)) {
+			parse_osmo_bts_features(bts, TLVP_VAL(tp, NM_ATT_MANUF_ID),
+						     TLVP_LEN(tp, NM_ATT_MANUF_ID));
 		}
-
-		/* check that max. expected BTS attribute is above given feature vector length */
-		if (len > OSMO_BYTES_FOR_BITS(_NUM_BTS_FEAT)) {
-			LOGPMO(&bts->mo, DNM, LOGL_NOTICE, "Get Attributes Response: reported unexpectedly long (%u bytes) "
-			       "feature vector - most likely it was compiled against newer BSC headers. "
-			       "Consider upgrading your BSC to later version.\n", len);
+		/* fall-through */
+	case GSM_BTS_TYPE_NANOBTS:
+		if (TLVP_PRESENT(tp, NM_ATT_IPACC_SUPP_FEATURES)) {
+			ipacc_parse_supp_features(bts, foh, TLVP_VAL(tp, NM_ATT_IPACC_SUPP_FEATURES),
+							    TLVP_LEN(tp, NM_ATT_IPACC_SUPP_FEATURES));
 		}
-
-		memcpy(bts->_features_data, TLVP_VAL(tp, NM_ATT_MANUF_ID), len);
-		bts->features_known = true;
-
-		/* Log each BTS feature in the reported vector */
-		for (i = 0; i < len * 8; i++) {
-			if (!osmo_bts_has_feature(&bts->features, i))
-				continue;
-
-			if (i >= _NUM_BTS_FEAT)
-				LOGPMO(&bts->mo, DNM, LOGL_NOTICE, "Get Attributes Response: unknown feature 0x%02x is"
-				       " supported\n", i);
-			else
-				LOGPMO(&bts->mo, DNM, LOGL_NOTICE, "Get Attributes Response: feature '%s' is"
-				       " supported\n", osmo_bts_features_name(i));
-		}
-
+		break;
+	default:
+		break;
 	}
 
 	/* Parse Attribute Response Info content for 3GPP TS 52.021 §9.4.28 Manufacturer Dependent State */
@@ -617,6 +640,7 @@ static int parse_attr_resp_info_attr(struct gsm_bts *bts, const struct gsm_bts_t
 
 	/* Parse Attribute Response Info content for 3GPP TS 52.021 §9.4.61 SW Configuration */
 	if (TLVP_PRESENT(tp, NM_ATT_SW_CONFIG)) {
+		struct abis_nm_sw_desc sw_descr[SW_DESCR_MAX];
 		data = TLVP_VAL(tp, NM_ATT_SW_CONFIG);
 		len = TLVP_LEN(tp, NM_ATT_SW_CONFIG);
 		/* after parsing manufacturer-specific attributes there's list of replies in form of sw-conf structure: */
@@ -728,15 +752,20 @@ static int abis_nm_rx_sw_act_req(struct msgb *mb)
 	struct tlv_parsed tp;
 	const uint8_t *sw_config;
 	int ret, sw_config_len, len;
-	struct abis_nm_sw_desc sw_descr[MAX_BTS_ATTR];
+	struct abis_nm_sw_desc sw_descr[SW_DESCR_MAX];
 
 	DEBUGPFOH(DNM, foh, "Software Activate Request, ACKing and Activating\n");
 
+	if (oh->length < sizeof(*foh)) {
+		LOGPFOH(DNM, LOGL_ERROR, foh, "Software Activate Request with length too small: %u\n", oh->length);
+		return -EINVAL;
+	}
+
 	ret = abis_nm_sw_act_req_ack(sign_link->trx->bts, foh->obj_class,
 				      foh->obj_inst.bts_nr,
 				      foh->obj_inst.trx_nr,
 				      foh->obj_inst.ts_nr, 0,
-				      foh->data, oh->length-sizeof(*foh));
+				      foh->data, oh->length - sizeof(*foh));
 	if (ret != 0) {
 		LOGPFOH(DNM, LOGL_ERROR, foh, "Sending SW ActReq ACK failed: %d\n", ret);
 		return ret;
@@ -1669,7 +1698,7 @@ int abis_nm_software_load(struct gsm_bts *bts, int trx_nr, const char *fname,
 		break;
 	case GSM_BTS_TYPE_NANOBTS:
 		sw->obj_class = NM_OC_BASEB_TRANSC;
-		sw->obj_instance[0] = sw->bts->nr;
+		sw->obj_instance[0] = sw->bts->bts_nr;
 		sw->obj_instance[1] = sw->trx_nr;
 		sw->obj_instance[2] = 0xff;
 		break;
@@ -2140,12 +2169,13 @@ int abis_nm_set_channel_attr(struct gsm_bts_trx_ts *ts, uint8_t chan_comb)
 }
 
 int abis_nm_sw_act_req_ack(struct gsm_bts *bts, uint8_t obj_class, uint8_t i1,
-			uint8_t i2, uint8_t i3, int nack, uint8_t *attr, int att_len)
+			   uint8_t i2, uint8_t i3, int nack,
+			   const uint8_t *attr, unsigned int attr_len)
 {
 	struct abis_om_hdr *oh;
 	struct msgb *msg = nm_msgb_alloc();
 	uint8_t msgtype = NM_MT_SW_ACT_REQ_ACK;
-	uint8_t len = att_len;
+	uint8_t len = attr_len;
 
 	if (nack) {
 		len += 2;
@@ -2153,12 +2183,10 @@ int abis_nm_sw_act_req_ack(struct gsm_bts *bts, uint8_t obj_class, uint8_t i1,
 	}
 
 	oh = (struct abis_om_hdr *) msgb_put(msg, ABIS_OM_FOM_HDR_SIZE);
-	fill_om_fom_hdr(oh, att_len, msgtype, obj_class, i1, i2, i3);
+	fill_om_fom_hdr(oh, len, msgtype, obj_class, i1, i2, i3);
 
-	if (attr) {
-		uint8_t *ptr = msgb_put(msg, att_len);
-		memcpy(ptr, attr, att_len);
-	}
+	if (attr != NULL && attr_len > 0)
+		memcpy(msgb_put(msg, attr_len), attr, attr_len);
 	if (nack)
 		msgb_tv_put(msg, NM_ATT_NACK_CAUSES, NM_NACK_OBJCLASS_NOTSUPP);
 
@@ -3078,7 +3106,7 @@ int abis_nm_ipaccess_restart(struct gsm_bts_trx *trx)
 
 	oh = (struct abis_om_hdr *) msgb_put(msg, ABIS_OM_FOM_HDR_SIZE);
 	fill_om_fom_hdr(oh, 0, NM_MT_IPACC_RESTART, NM_OC_BASEB_TRANSC,
-			trx->bts->nr, trx->nr, 0xff);
+			trx->bts->bts_nr, trx->nr, 0xff);
 
 	return abis_nm_sendmsg_direct(trx->bts, msg);
 }
