@@ -24,9 +24,6 @@
 
 #include <osmocom/ctrl/control_cmd.h>
 
-#include <osmocom/vty/command.h>
-#include <osmocom/vty/misc.h>
-
 #include <osmocom/bsc/ctrl.h>
 #include <osmocom/bsc/osmo_bsc_rf.h>
 #include <osmocom/bsc/bts.h>
@@ -252,6 +249,8 @@ CTRL_CMD_DEFINE_RANGE(bts_lac, "location-area-code", struct gsm_bts, location_ar
 CTRL_CMD_DEFINE_RANGE(bts_ci, "cell-identity", struct gsm_bts, cell_identity, 0, 65535);
 CTRL_CMD_DEFINE_RANGE(bts_bsic, "bsic", struct gsm_bts, bsic, 0, 63);
 CTRL_CMD_DEFINE_RANGE(bts_rach_max_delay, "rach-max-delay", struct gsm_bts, rach_max_delay, 1, 127);
+CTRL_CMD_DEFINE_RANGE(bts_rach_expiry_timeout, "rach-expiry-timeout", struct gsm_bts, rach_expiry_timeout, 4, 64);
+CTRL_CMD_DEFINE_RANGE(bts_ms_max_power, "ms-max-power", struct gsm_bts, ms_max_power, 0, 40);
 
 static int set_bts_apply_config(struct ctrl_cmd *cmd, void *data)
 {
@@ -370,7 +369,7 @@ static int get_bts_oml_up(struct ctrl_cmd *cmd, void *data)
 {
 	const struct gsm_bts *bts = cmd->node;
 
-	cmd->reply = talloc_asprintf(cmd, "%llu", bts_uptime(bts));
+	cmd->reply = talloc_asprintf(cmd, "%llu", bts->oml_link ? bts_updowntime(bts) : 0);
 	if (!cmd->reply) {
 		cmd->reply = "OOM";
 		return CTRL_CMD_ERROR;
@@ -717,7 +716,6 @@ CTRL_CMD_DEFINE_WO(bts_neighbor_list_mode, "neighbor-list mode");
 static int set_bts_si2quater_neighbor_list_del_earfcn(struct ctrl_cmd *cmd, void *data)
 {
 	struct gsm_bts *bts = (struct gsm_bts *)cmd->node;
-	struct osmo_earfcn_si2q *e = &bts->si_common.si2quater_neigh_list;
 	int earfcn;
 
 	if (osmo_str_to_int(&earfcn, cmd->value, 10, 0, 65535) < 0) {
@@ -725,7 +723,7 @@ static int set_bts_si2quater_neighbor_list_del_earfcn(struct ctrl_cmd *cmd, void
 		return CTRL_CMD_ERROR;
 	}
 
-	if (osmo_earfcn_del(e, earfcn) < 0) {
+	if (bts_earfcn_del(bts, earfcn) < 0) {
 		cmd->reply = "Failed to delete a (not existent?) neighbor EARFCN";
 		return CTRL_CMD_ERROR;
 	}
@@ -835,7 +833,6 @@ static int verify_bts_si2quater_neighbor_list_add_earfcn(struct ctrl_cmd *cmd, c
 static int set_bts_si2quater_neighbor_list_add_earfcn(struct ctrl_cmd *cmd, void *data)
 {
 	struct gsm_bts *bts = (struct gsm_bts *)cmd->node;
-	struct osmo_earfcn_si2q *neighbors = &bts->si_common.si2quater_neigh_list;
 	char *earfcn_str, *thresh_hi_str, *thresh_lo_str, *prio_str, *qrxlv_str, *meas_str, *saveptr, *tmp;
 	int earfcn, thresh_hi, thresh_lo, prio, qrxlv, meas, result;
 
@@ -914,7 +911,7 @@ static int set_bts_si2quater_neighbor_list_add_earfcn(struct ctrl_cmd *cmd, void
 			cmd->reply = "OOM";
 	}
 
-	if (osmo_earfcn_del(neighbors, earfcn) != 0)
+	if (bts_earfcn_del(bts, earfcn) != 0)
 		cmd->reply = "Failed to roll-back adding EARFCN";
 
 	return CTRL_CMD_ERROR;
@@ -1041,6 +1038,7 @@ static int set_bts_cell_reselection_offset(struct ctrl_cmd *cmd, void *data)
 	struct gsm_bts *bts = cmd->node;
 	bts->si_common.cell_ro_sel_par.present = 1;
 	bts->si_common.cell_ro_sel_par.cell_resel_off = atoi(cmd->value) / 2;
+	cmd->reply = "OK";
 	return CTRL_CMD_REPLY;
 }
 
@@ -1145,6 +1143,91 @@ static int set_bts_cell_reselection_hysteresis(struct ctrl_cmd *cmd, void *data)
 
 CTRL_CMD_DEFINE(bts_cell_reselection_hysteresis, "cell-reselection-hysteresis");
 
+
+static int verify_bts_radio_link_timeout(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	int radio_link_timeout;
+	struct gsm_bts *bts = cmd->node;
+
+	if (strcmp(value, "infinite") == 0) {
+		if (bts->type != GSM_BTS_TYPE_OSMOBTS) {
+			cmd->reply = "Infinite radio link timeout not supported by BTS";
+			return 1;
+		}
+		return 0;
+	}
+
+	radio_link_timeout = atoi(cmd->value);
+
+	if (radio_link_timeout < 0 || radio_link_timeout > 64) {
+		cmd->reply = "Value is out of range";
+		return 1;
+	} else if (radio_link_timeout % 4 != 0) {
+		cmd->reply = "Value must be a multiple of 4";
+		return 1;
+	}
+
+	return 0;
+}
+
+static int get_bts_radio_link_timeout(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+
+	cmd->reply = talloc_asprintf(cmd, "%u", gsm_bts_get_radio_link_timeout(bts));
+	if (!cmd->reply) {
+		cmd->reply = "OOM";
+		return CTRL_CMD_ERROR;
+	}
+
+	return CTRL_CMD_REPLY;
+}
+
+static int set_bts_radio_link_timeout(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+	gsm_bts_set_radio_link_timeout(bts, atoi(cmd->value));
+	cmd->reply = "OK";
+	return CTRL_CMD_REPLY;
+}
+
+CTRL_CMD_DEFINE(bts_radio_link_timeout, "radio-link-timeout");
+
+static int verify_bts_rxlev_access_min(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	int rxlev_access_min = atoi(cmd->value);
+
+	if (rxlev_access_min < 0 || rxlev_access_min > 63) {
+		cmd->reply = "Value is out of range";
+		return 1;
+	}
+
+	return 0;
+}
+
+static int get_bts_rxlev_access_min(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+
+	cmd->reply = talloc_asprintf(cmd, "%u", bts->si_common.cell_sel_par.rxlev_acc_min);
+	if (!cmd->reply) {
+		cmd->reply = "OOM";
+		return CTRL_CMD_ERROR;
+	}
+
+	return CTRL_CMD_REPLY;
+}
+
+static int set_bts_rxlev_access_min(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+	bts->si_common.cell_sel_par.rxlev_acc_min = atoi(cmd->value);
+	cmd->reply = "OK";
+	return CTRL_CMD_REPLY;
+}
+
+CTRL_CMD_DEFINE(bts_rxlev_access_min, "rach-rxlev-access-min");
+
 /* Return space concatenated set of pairs <class>,<barred/allowed> */
 static int get_bts_rach_access_control_class(struct ctrl_cmd *cmd, void *data)
 {
@@ -1264,6 +1347,72 @@ static int set_bts_rach_access_control_class_allow(struct ctrl_cmd *cmd, void *d
 
 CTRL_CMD_DEFINE_WO(bts_rach_access_control_class_allow, "rach-access-control-class allow");
 
+static int verify_bts_rach_cell_barred(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	int bar = atoi(cmd->value);
+
+	if ((bar != 0) && (bar != 1))
+		return 1;
+
+	return 0;
+}
+
+static int get_bts_rach_cell_barred(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+
+	cmd->reply = talloc_asprintf(cmd, "%u", bts->si_common.rach_control.cell_bar);
+	if (!cmd->reply) {
+		cmd->reply = "OOM";
+		return CTRL_CMD_ERROR;
+	}
+
+	return CTRL_CMD_REPLY;
+}
+
+static int set_bts_rach_cell_barred(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+	bts->si_common.rach_control.cell_bar = atoi(cmd->value);
+	cmd->reply = "OK";
+	return CTRL_CMD_REPLY;
+}
+
+CTRL_CMD_DEFINE(bts_rach_cell_barred, "rach-cell-barred");
+
+static int verify_bts_rach_max_trans(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	int max_trans = atoi(cmd->value);
+
+	if ((max_trans != 1) && (max_trans != 2) && (max_trans != 4) && (max_trans != 7))
+		return 1;
+
+	return 0;
+}
+
+static int get_bts_rach_max_trans(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+
+	cmd->reply = talloc_asprintf(cmd, "%u", rach_max_trans_raw2val(bts->si_common.rach_control.max_trans));
+	if (!cmd->reply) {
+		cmd->reply = "OOM";
+		return CTRL_CMD_ERROR;
+	}
+
+	return CTRL_CMD_REPLY;
+}
+
+static int set_bts_rach_max_trans(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+	bts->si_common.rach_control.max_trans = rach_max_trans_val2raw(atoi(cmd->value));
+	cmd->reply = "OK";
+	return CTRL_CMD_REPLY;
+}
+
+CTRL_CMD_DEFINE(bts_rach_max_trans, "rach-max-transmission");
+
 /* Return space concatenated set of tuples <UARFCN>,<scrambling code>,<diversity bit> */
 static int get_bts_neighbor_list_si2quater_uarfcn(struct ctrl_cmd *cmd, void *data)
 {
@@ -1330,6 +1479,53 @@ static int get_bts_neighbor_list_si2quater_earfcn(struct ctrl_cmd *cmd, void *da
 
 CTRL_CMD_DEFINE_RO(bts_neighbor_list_si2quater_earfcn, "neighbor-list si2quater earfcns");
 
+char *bts_lchan_dump_full_ctrl(const void *t, struct gsm_bts *bts)
+{
+	int trx_nr;
+	bool first_trx = true;
+	char *trx_dump, *dump;
+	struct gsm_bts_trx *trx;
+
+	dump = talloc_strdup(t, "");
+	if (!dump)
+		return NULL;
+
+	for (trx_nr = 0; trx_nr < bts->num_trx; trx_nr++) {
+		trx = gsm_bts_trx_num(bts, trx_nr);
+		trx_dump = trx_lchan_dump_full_ctrl(t, trx);
+		if (!trx_dump)
+			return NULL;
+		if (!strlen(trx_dump))
+			continue;
+		dump = talloc_asprintf_append(dump, first_trx ? "%s" : "\n%s", trx_dump);
+		if (!dump)
+			return NULL;
+		first_trx = false;
+	}
+
+	return dump;
+}
+
+/* Return full information about all logical channels in a BTS.
+ * format: bts.<0-255>.show-lchan.full
+ * result format: New line delimited list of <bts>,<trx>,<ts>,<lchan>,<type>,<connection>,<state>,<last error>,<bs power>,
+ *  <ms power>,<interference dbm>, <interference band>,<channel mode>,<imsi>,<tmsi>,<ipa bound ip>,<ipa bound port>,
+ *  <ipa bound conn id>,<ipa conn ip>,<ipa conn port>,<ipa conn speech mode>
+ */
+static int get_bts_show_lchan_full(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+
+	cmd->reply = bts_lchan_dump_full_ctrl(cmd, bts);
+	if (!cmd->reply) {
+		cmd->reply = "OOM";
+		return CTRL_CMD_ERROR;
+	}
+
+	return CTRL_CMD_REPLY;
+}
+CTRL_CMD_DEFINE_RO(bts_show_lchan_full, "show-lchan full");
+
 int bsc_bts_ctrl_cmds_install(void)
 {
 	int rc = 0;
@@ -1339,6 +1535,7 @@ int bsc_bts_ctrl_cmds_install(void)
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_ci);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_bsic);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rach_max_delay);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rach_expiry_timeout);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_apply_config);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_si);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_power_ctrl_defs);
@@ -1363,11 +1560,17 @@ int bsc_bts_ctrl_cmds_install(void)
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_cell_reselection_offset);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_cell_reselection_penalty_time);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_cell_reselection_hysteresis);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_ms_max_power);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_radio_link_timeout);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rxlev_access_min);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rach_access_control_class);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rach_access_control_class_bar);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rach_access_control_class_allow);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rach_cell_barred);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rach_max_trans);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_neighbor_list_si2quater_uarfcn);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_neighbor_list_si2quater_earfcn);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_show_lchan_full);
 
 	rc |= neighbor_ident_ctrl_init();
 
