@@ -195,156 +195,78 @@ static int nm_statechg_event(int evt, struct nm_statechg_signal_data *nsd)
 /* Callback function to be called every time we receive a 12.21 SW activated report */
 static int sw_activ_rep(struct msgb *mb)
 {
-	struct abis_om_fom_hdr *foh = msgb_l3(mb);
+	const struct abis_om_fom_hdr *foh = msgb_l3(mb);
 	struct e1inp_sign_link *sign_link = mb->dst;
 	struct gsm_bts *bts = sign_link->trx->bts;
-	struct gsm_bts_trx *trx;
-	struct gsm_gprs_nsvc *nsvc;
-	struct gsm_bts_trx_ts *ts;
+	struct gsm_abis_mo *mo;
+	struct tlv_parsed tp;
 
 	if (!is_ipa_abisip_bts(bts))
 		return 0;
 
-	switch (foh->obj_class) {
-	case NM_OC_SITE_MANAGER:
-		osmo_fsm_inst_dispatch(bts->site_mgr->mo.fi, NM_EV_SW_ACT_REP, NULL);
-		break;
-	case NM_OC_BTS:
-		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_SW_ACT_REP, NULL);
-		break;
-	case NM_OC_BASEB_TRANSC:
-		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
-			return -EINVAL;
-		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_SW_ACT_REP, NULL);
-		break;
-	case NM_OC_RADIO_CARRIER:
-		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
-			return -EINVAL;
-		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_SW_ACT_REP, NULL);
-		break;
-	case NM_OC_CHANNEL:
-		if (!(ts = abis_nm_get_ts(mb)))
-			return -EINVAL;
-		osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_SW_ACT_REP, NULL);
-		break;
-	case NM_OC_GPRS_NSE:
-		osmo_fsm_inst_dispatch(bts->site_mgr->gprs.nse.mo.fi, NM_EV_SW_ACT_REP, NULL);
-		break;
-	case NM_OC_GPRS_CELL:
-		osmo_fsm_inst_dispatch(bts->gprs.cell.mo.fi, NM_EV_SW_ACT_REP, NULL);
-		break;
-	case NM_OC_GPRS_NSVC:
-		if (!(nsvc = gsm_bts_sm_nsvc_num(bts->site_mgr, foh->obj_inst.trx_nr)))
-			return -EINVAL;
-		osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_SW_ACT_REP, NULL);
-		break;
+	mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
+	if (mo == NULL) {
+		LOGPFOH(DNM, LOGL_ERROR, foh, "Rx SW activated report for non-existent MO\n");
+		return -ENOENT;
 	}
-	return 0;
-}
 
-static void nm_rx_opstart_ack_chan(struct msgb *oml_msg)
-{
-	struct gsm_bts_trx_ts *ts;
-	ts = abis_nm_get_ts(oml_msg);
-	if (!ts)
-		/* error already logged in abis_nm_get_ts() */
-		return;
-	if (!ts->fi) {
-		LOG_TS(ts, LOGL_ERROR, "Channel OPSTART ACK for uninitialized TS\n");
-		return;
+	if (abis_nm_tlv_parse(&tp, bts, &foh->data[0], msgb_l3len(mb) - sizeof(*foh)) < 0) {
+		LOGPFOH(DNM, LOGL_ERROR, foh, "%s(): tlv_parse failed\n", __func__);
+		return -EINVAL;
 	}
-	osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_OPSTART_ACK, NULL);
+
+	mo->ipaccess.obj_version = 0; /* implicit default */
+	if (TLVP_PRES_LEN(&tp, NM_ATT_IPACC_OBJ_VERSION, 1)) {
+		const uint8_t *versions = TLVP_VAL(&tp, NM_ATT_IPACC_OBJ_VERSION);
+		char buf[256];
+		struct osmo_strbuf sb = { .buf = buf, .len = sizeof(buf) };
+
+		/* nanoBTS may report several Object Versions;  the first one will
+		 * be used by default unless requested explicitly before OPSTARTing. */
+		mo->ipaccess.obj_version = versions[0];
+
+		OSMO_STRBUF_PRINTF(sb, "%u (default)", versions[0]);
+		for (uint16_t i = 1; i < TLVP_LEN(&tp, NM_ATT_IPACC_OBJ_VERSION); i++)
+			OSMO_STRBUF_PRINTF(sb, ", %u", versions[i]);
+		LOGPFOH(DNM, LOGL_INFO, foh, "IPA Object Versions supported: %s\n", buf);
+	}
+
+	osmo_fsm_inst_dispatch(mo->fi, NM_EV_SW_ACT_REP, NULL);
+	return 0;
 }
 
 static void nm_rx_opstart_ack(struct msgb *oml_msg)
 {
-	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
+	const struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
 	struct e1inp_sign_link *sign_link = oml_msg->dst;
 	struct gsm_bts *bts = sign_link->trx->bts;
-	struct gsm_bts_trx *trx;
-	struct gsm_gprs_nsvc *nsvc;
+	struct gsm_abis_mo *mo;
 
-	switch (foh->obj_class) {
-	case NM_OC_SITE_MANAGER:
-		osmo_fsm_inst_dispatch(bts->site_mgr->mo.fi, NM_EV_OPSTART_ACK, NULL);
-		break;
-	case NM_OC_BTS:
-		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_OPSTART_ACK, NULL);
-		break;
-	case NM_OC_RADIO_CARRIER:
-		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
-			return;
-		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_OPSTART_ACK, NULL);
-		break;
-	case NM_OC_BASEB_TRANSC:
-		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
-			return;
-		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_OPSTART_ACK, NULL);
-		break;
-	case NM_OC_CHANNEL:
-		nm_rx_opstart_ack_chan(oml_msg);
-		break;
-	case NM_OC_GPRS_NSE:
-		osmo_fsm_inst_dispatch(bts->site_mgr->gprs.nse.mo.fi, NM_EV_OPSTART_ACK, NULL);
-		break;
-	case NM_OC_GPRS_CELL:
-		osmo_fsm_inst_dispatch(bts->gprs.cell.mo.fi, NM_EV_OPSTART_ACK, NULL);
-		break;
-	case NM_OC_GPRS_NSVC:
-		if (!(nsvc = gsm_bts_sm_nsvc_num(bts->site_mgr, foh->obj_inst.trx_nr)))
-			return;
-		osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_OPSTART_ACK, NULL);
-		break;
-	default:
-		break;
-	}
+	if (!is_ipa_abisip_bts(bts))
+		return;
+
+	mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
+	if (mo == NULL)
+		LOGPFOH(DNM, LOGL_ERROR, foh, "Rx OPSTART ACK for non-existent MO\n");
+	else
+		osmo_fsm_inst_dispatch(mo->fi, NM_EV_OPSTART_ACK, NULL);
 }
 
 static void nm_rx_opstart_nack(struct msgb *oml_msg)
 {
-	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
+	const struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
 	struct e1inp_sign_link *sign_link = oml_msg->dst;
 	struct gsm_bts *bts = sign_link->trx->bts;
-	struct gsm_bts_trx *trx;
-	struct gsm_bts_trx_ts *ts;
-	struct gsm_gprs_nsvc *nsvc;
+	struct gsm_abis_mo *mo;
 
-	switch (foh->obj_class) {
-	case NM_OC_SITE_MANAGER:
-		osmo_fsm_inst_dispatch(bts->site_mgr->mo.fi, NM_EV_OPSTART_NACK, NULL);
-		break;
-	case NM_OC_BTS:
-		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_OPSTART_ACK, NULL);
-		break;
-	case NM_OC_RADIO_CARRIER:
-		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
-			return;
-		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_OPSTART_NACK, NULL);
-		break;
-	case NM_OC_BASEB_TRANSC:
-		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
-			return;
-		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_OPSTART_NACK, NULL);
-		break;
-	case NM_OC_CHANNEL:
-		if (!(ts = abis_nm_get_ts(oml_msg)))
-			return;
-		osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_OPSTART_NACK, NULL);
-		break;
-	case NM_OC_GPRS_NSE:
-		osmo_fsm_inst_dispatch(bts->site_mgr->gprs.nse.mo.fi, NM_EV_OPSTART_NACK, NULL);
-		break;
-	case NM_OC_GPRS_CELL:
-		osmo_fsm_inst_dispatch(bts->gprs.cell.mo.fi, NM_EV_OPSTART_NACK, NULL);
-		break;
-	case NM_OC_GPRS_NSVC:
-		if (!(nsvc = gsm_bts_sm_nsvc_num(bts->site_mgr, foh->obj_inst.trx_nr)))
-			return;
-		osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_OPSTART_NACK, NULL);
-		break;
-	default:
-		break;
-	}
+	if (!is_ipa_abisip_bts(bts))
+		return;
+
+	mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
+	if (mo == NULL)
+		LOGPFOH(DNM, LOGL_ERROR, foh, "Rx OPSTART NACK for non-existent MO\n");
+	else
+		osmo_fsm_inst_dispatch(mo->fi, NM_EV_OPSTART_NACK, NULL);
 }
 
 static void nm_rx_get_attr_rep(struct msgb *oml_msg)
@@ -352,20 +274,16 @@ static void nm_rx_get_attr_rep(struct msgb *oml_msg)
 	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
 	struct e1inp_sign_link *sign_link = oml_msg->dst;
 	struct gsm_bts *bts = sign_link->trx->bts;
-	struct gsm_bts_trx *trx;
+	struct gsm_abis_mo *mo;
 
-	switch (foh->obj_class) {
-	case NM_OC_BTS:
-		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_GET_ATTR_REP, NULL);
-		break;
-	case NM_OC_BASEB_TRANSC:
-		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
-			return;
-		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_GET_ATTR_REP, NULL);
-		break;
-	default:
-		LOGPFOH(DNM, LOGL_ERROR, foh, "Get Attributes Response received on incorrect object class %d!\n", foh->obj_class);
-	}
+	if (!is_ipa_abisip_bts(bts))
+		return;
+
+	mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
+	if (mo == NULL)
+		LOGPFOH(DNM, LOGL_ERROR, foh, "Rx Get Attribute Report for non-existent MO\n");
+	else
+		osmo_fsm_inst_dispatch(mo->fi, NM_EV_GET_ATTR_REP, NULL);
 }
 
 static void nm_rx_set_bts_attr_ack(struct msgb *oml_msg)
@@ -374,6 +292,9 @@ static void nm_rx_set_bts_attr_ack(struct msgb *oml_msg)
 	struct e1inp_sign_link *sign_link = oml_msg->dst;
 	struct gsm_bts *bts = sign_link->trx->bts;
 
+	if (!is_ipa_abisip_bts(bts))
+		return;
+
 	if (foh->obj_class != NM_OC_BTS) {
 		LOG_BTS(bts, DNM, LOGL_ERROR, "Set BTS Attr Ack received on non BTS object!\n");
 		return;
@@ -387,8 +308,12 @@ static void nm_rx_set_radio_attr_ack(struct msgb *oml_msg)
 	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
 	struct e1inp_sign_link *sign_link = oml_msg->dst;
 	struct gsm_bts *bts = sign_link->trx->bts;
-	struct gsm_bts_trx *trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr);
+	struct gsm_bts_trx *trx;
 
+	if (!is_ipa_abisip_bts(bts))
+		return;
+
+	trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr);
 	if (!trx || foh->obj_class != NM_OC_RADIO_CARRIER) {
 		LOGPFOH(DNM, LOGL_ERROR, foh, "Set Radio Carrier Attr Ack received on non Radio Carrier object!\n");
 		return;
@@ -399,8 +324,14 @@ static void nm_rx_set_radio_attr_ack(struct msgb *oml_msg)
 static void nm_rx_set_chan_attr_ack(struct msgb *oml_msg)
 {
 	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
-	struct gsm_bts_trx_ts *ts = abis_nm_get_ts(oml_msg);
+	struct e1inp_sign_link *sign_link = oml_msg->dst;
+	struct gsm_bts *bts = sign_link->trx->bts;
+	struct gsm_bts_trx_ts *ts;
 
+	if (!is_ipa_abisip_bts(bts))
+		return;
+
+	ts = abis_nm_get_ts(oml_msg);
 	if (!ts || foh->obj_class != NM_OC_CHANNEL) {
 		LOGPFOH(DNM, LOGL_ERROR, foh, "Set Channel Attr Ack received on non Radio Channel object!\n");
 		return;
@@ -596,6 +527,8 @@ void ipaccess_drop_oml(struct gsm_bts *bts, const char *reason)
 	struct gsm_bts_trx_ts *ts ;
 	uint8_t tn;
 	uint8_t i;
+	struct timespec tp;
+	int rc;
 
 	/* First of all, remove deferred drop if enabled */
 	osmo_timer_del(&bts->oml_drop_link_timer);
@@ -606,7 +539,8 @@ void ipaccess_drop_oml(struct gsm_bts *bts, const char *reason)
 	LOG_BTS(bts, DLINP, LOGL_NOTICE, "Dropping OML link: %s\n", reason);
 	e1inp_sign_link_destroy(bts->oml_link);
 	bts->oml_link = NULL;
-	bts->uptime = 0;
+	rc = osmo_clock_gettime(CLOCK_MONOTONIC, &tp);
+	bts->updowntime = (rc < 0) ? 0 : tp.tv_sec; /* we don't need sub-second precision for downtime */
 	osmo_stat_item_dec(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_OML_CONNECTED), 1);
 	gsm_bts_stats_reset(bts);
 
@@ -779,7 +713,7 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 						E1INP_SIGN_OML, bts->c0,
 						bts->oml_tei, 0);
 		rc = osmo_clock_gettime(CLOCK_MONOTONIC, &tp);
-		bts->uptime = (rc < 0) ? 0 : tp.tv_sec; /* we don't need sub-second precision for uptime */
+		bts->updowntime = (rc < 0) ? 0 : tp.tv_sec; /* we don't need sub-second precision for uptime */
 		if (!(sign_link->trx->bts->ip_access.flags & OML_UP)) {
 			e1inp_event(sign_link->ts, S_L_INP_TEI_UP,
 					sign_link->tei, sign_link->sapi);
