@@ -29,8 +29,10 @@
 #include <osmocom/bsc/bsc_subscriber.h>
 #include <osmocom/bsc/paging.h>
 #include <osmocom/bsc/gsm_04_08_rr.h>
+#include <osmocom/bsc/gsm_08_08.h>
 #include <osmocom/bsc/bsc_subscr_conn_fsm.h>
 #include <osmocom/bsc/codec_pref.h>
+#include <osmocom/bsc/data_rate_pref.h>
 #include <osmocom/bsc/abis_rsl.h>
 #include <osmocom/bsc/handover_fsm.h>
 #include <osmocom/bsc/bts.h>
@@ -47,6 +49,8 @@
 #include <osmocom/core/sockaddr_str.h>
 #include <osmocom/bsc/lcs_loc_req.h>
 #include <osmocom/bsc/bssmap_reset.h>
+#include <osmocom/bsc/assignment_fsm.h>
+#include <osmocom/bsc/vgcs_fsm.h>
 
 #define IP_V4_ADDR_LEN 4
 
@@ -350,8 +354,12 @@ int bsc_paging_start(struct bsc_paging_params *params)
 			return -EINVAL;
 		}
 	}
-	if (params->tmsi != GSM_RESERVED_TMSI)
-		params->bsub->tmsi = params->tmsi;
+	if (params->tmsi != GSM_RESERVED_TMSI) {
+		if (bsc_subscr_set_tmsi(params->bsub, params->tmsi) < 0) {
+			LOG_PAGING(params, DMSC, LOGL_ERROR, "Paging request failed: Could not set TMSI on subscriber\n");
+			return -EINVAL;
+		}
+	}
 	log_set_context(LOG_CTX_BSC_SUBSCR, params->bsub);
 
 	switch (params->cil.id_discr) {
@@ -667,19 +675,79 @@ static int bssmap_handle_lcls_connect_ctrl(struct gsm_subscriber_connection *con
 	return 0;
 }
 
-/* Select a preferred and an alternative codec rate depending on the available
+/* Select a preferred and an alternative data rate depending on the available
  * capabilities. This decision does not include the actual channel load yet,
  * this is also the reason why the result is a preferred and an alternate
  * setting. The final decision is made in assignment_fsm.c when the actual
  * lchan is requested. The preferred lchan will be requested first. If we
  * find an alternate setting here, this one will be tried secondly if our
  * primary choice fails. */
-static int select_codecs(struct assignment_request *req, struct gsm0808_channel_type *ct,
-			 struct gsm_subscriber_connection *conn)
+static int select_data_rates(struct assignment_request *req, struct gsm0808_channel_type *ct,
+			     struct gsm_subscriber_connection *conn)
+{
+	int rc, i, nc = 0;
+
+	switch (ct->ch_rate_type) {
+	case GSM0808_DATA_FULL_BM:
+		rc = match_data_rate_pref(&req->ch_mode_rate_list[nc], ct, true);
+		nc += (rc == 0) ? 1 : 0;
+		break;
+	case GSM0808_DATA_HALF_LM:
+		rc = match_data_rate_pref(&req->ch_mode_rate_list[nc], ct, false);
+		nc += (rc == 0) ? 1 : 0;
+		break;
+	case GSM0808_DATA_FULL_PREF_NO_CHANGE:
+	case GSM0808_DATA_FULL_PREF:
+		rc = match_data_rate_pref(&req->ch_mode_rate_list[nc], ct, true);
+		nc += (rc == 0) ? 1 : 0;
+		rc = match_data_rate_pref(&req->ch_mode_rate_list[nc], ct, false);
+		nc += (rc == 0) ? 1 : 0;
+		break;
+	case GSM0808_DATA_HALF_PREF_NO_CHANGE:
+	case GSM0808_DATA_HALF_PREF:
+		rc = match_data_rate_pref(&req->ch_mode_rate_list[nc], ct, false);
+		nc += (rc == 0) ? 1 : 0;
+		rc = match_data_rate_pref(&req->ch_mode_rate_list[nc], ct, true);
+		nc += (rc == 0) ? 1 : 0;
+		break;
+	default:
+		rc = -EINVAL;
+		break;
+	}
+
+	if (!nc) {
+		LOGP(DMSC, LOGL_ERROR, "No supported data rate found for channel_type ="
+		     " { ch_indctr=0x%x, ch_rate_type=0x%x, perm_spch=[%s] }\n",
+		     ct->ch_indctr, ct->ch_rate_type, osmo_hexdump(ct->perm_spch, ct->perm_spch_len));
+		return -EINVAL;
+	}
+
+	for (i = 0; i < nc; i++) {
+		DEBUGP(DMSC, "Found matching data rate (pref=%d): %s %s for channel_type ="
+		       " { ch_indctr=0x%x, ch_rate_type=0x%x, perm_spch=[ %s] }\n",
+		       i,
+		       req->ch_mode_rate_list[i].chan_rate == CH_RATE_FULL ? "full rate" : "half rate",
+		       get_value_string(gsm48_chan_mode_names, req->ch_mode_rate_list[i].chan_mode),
+		       ct->ch_indctr, ct->ch_rate_type, osmo_hexdump(ct->perm_spch, ct->perm_spch_len));
+	}
+
+	req->n_ch_mode_rate = nc;
+
+	return 0;
+}
+
+/* Select a preferred and an alternative codec rate depending on the available
+ * capabilities. This decision does not include the actual lchan availability yet,
+ * this is also the reason why the result is a preferred and an alternate
+ * setting. The final decision is made in assignment_fsm.c when the actual
+ * lchan is requested. The preferred lchan will be requested first. If we
+ * find an alternate setting here, this one will be tried secondly if our
+ * primary choice fails. */
+static int select_codecs(struct assignment_request *req, const struct gsm0808_channel_type *ct,
+			 struct gsm_subscriber_connection *conn, struct gsm_bts *bts)
 {
 	int rc, i, nc = 0;
 	struct bsc_msc_data *msc;
-	struct gsm_bts *bts = conn_get_bts(conn);
 
 	if (!bts) {
 		LOGP(DMSC, LOGL_ERROR, "No lchan, cannot select codecs\n");
@@ -796,6 +864,244 @@ static int select_sign_chan(struct assignment_request *req, struct gsm0808_chann
 	return nc > 0 ? 0 : -EINVAL;
 }
 
+static int bssmap_handle_ass_req_tp_cic(struct tlv_parsed *tp, bool aoip, uint16_t *cic, uint8_t *cause)
+{
+	if (TLVP_PRESENT(tp, GSM0808_IE_CIRCUIT_IDENTITY_CODE)) {
+		/* CIC is permitted in both AoIP and SCCPlite */
+		*cic = osmo_load16be(TLVP_VAL(tp, GSM0808_IE_CIRCUIT_IDENTITY_CODE));
+		return 0;
+	}
+
+	if (!aoip) {
+		/* no CIC but SCCPlite: illegal */
+		LOGP(DMSC, LOGL_ERROR, "SCCPlite MSC, but no CIC in ASSIGN REQ?\n");
+		*cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
+		return -1;
+	}
+
+	return 0;
+}
+
+static int bssmap_handle_ass_req_tp_rtp_addr(struct tlv_parsed *tp, bool aoip, char *msc_rtp_addr,
+					     size_t msc_rtp_addr_len, uint16_t *msc_rtp_port, uint8_t *cause)
+{
+	struct sockaddr_storage rtp_addr;
+	int rc;
+	unsigned int rc2;
+
+	if (TLVP_PRESENT(tp, GSM0808_IE_AOIP_TRASP_ADDR)) {
+		if (!aoip) {
+			/* SCCPlite and AoIP transport address: illegal */
+			LOGP(DMSC, LOGL_ERROR, "AoIP Transport address over IPA ?!?\n");
+			*cause = GSM0808_CAUSE_INCORRECT_VALUE;
+			return -1;
+		}
+		/* Decode AoIP transport address element */
+		rc = gsm0808_dec_aoip_trasp_addr(&rtp_addr,
+						 TLVP_VAL(tp, GSM0808_IE_AOIP_TRASP_ADDR),
+						 TLVP_LEN(tp, GSM0808_IE_AOIP_TRASP_ADDR));
+		if (rc < 0) {
+			LOGP(DMSC, LOGL_ERROR, "Unable to decode AoIP transport address.\n");
+			*cause = GSM0808_CAUSE_INCORRECT_VALUE;
+			return -1;
+		}
+
+		rc2 = osmo_sockaddr_to_str_and_uint(msc_rtp_addr, msc_rtp_addr_len, msc_rtp_port,
+						    (const struct sockaddr *)&rtp_addr);
+		if (!rc2 || rc >= msc_rtp_addr_len) {
+			LOGP(DMSC, LOGL_ERROR, "Assignment request: RTP address is too long\n");
+			*cause = GSM0808_CAUSE_REQ_CODEC_TYPE_OR_CONFIG_UNAVAIL;
+			return -1;
+		}
+		return 0;
+	}
+
+	if (aoip) {
+		/* no AoIP transport level address but AoIP transport: illegal */
+		LOGP(DMSC, LOGL_ERROR, "AoIP transport address missing in ASSIGN REQ, "
+		     "audio would not work; rejecting\n");
+		*cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
+		return -1;
+	}
+
+	return 0;
+}
+
+static int bssmap_handle_ass_req_tp_osmux(struct gsm_subscriber_connection *conn, struct tlv_parsed *tp,
+					  bool *use_osmux, uint8_t *osmux_cid, uint8_t *cause)
+{
+	int rc;
+
+	if (TLVP_PRESENT(tp, GSM0808_IE_OSMO_OSMUX_CID)) {
+		if (conn->sccp.msc->use_osmux == OSMUX_USAGE_OFF) {
+			LOGP(DMSC, LOGL_ERROR, "MSC using Osmux but we have it disabled.\n");
+			*cause = GSM0808_CAUSE_INCORRECT_VALUE;
+			return -1;
+		}
+		*use_osmux = true;
+		rc = gsm0808_dec_osmux_cid(osmux_cid,
+					   TLVP_VAL(tp, GSM0808_IE_OSMO_OSMUX_CID),
+					   TLVP_LEN(tp, GSM0808_IE_OSMO_OSMUX_CID));
+		if (rc < 0) {
+			LOGP(DMSC, LOGL_ERROR, "Unable to decode Osmux CID.\n");
+			*cause = GSM0808_CAUSE_INCORRECT_VALUE;
+			return -1;
+		}
+		return 0;
+	}
+
+	if (conn->sccp.msc->use_osmux == OSMUX_USAGE_ONLY) {
+		LOGP(DMSC, LOGL_ERROR, "MSC not using Osmux but we are forced to use it.\n");
+		*cause = GSM0808_CAUSE_INCORRECT_VALUE;
+		return -1;
+	}
+
+	if (conn->sccp.msc->use_osmux == OSMUX_USAGE_ON)
+		LOGP(DMSC, LOGL_NOTICE, "MSC not using Osmux but we have Osmux enabled.\n");
+
+	return 0;
+}
+
+static int bssmap_handle_ass_req_tp_codec_list(struct gsm_subscriber_connection *conn, struct tlv_parsed *tp, bool aoip,
+					       uint8_t *cause)
+{
+	int rc;
+
+	/* Decode speech codec list. First set len = 0. */
+	conn->codec_list = (struct gsm0808_speech_codec_list){};
+	/* Check for speech codec list element */
+	if (TLVP_PRESENT(tp, GSM0808_IE_SPEECH_CODEC_LIST)) {
+		/* Decode Speech Codec list */
+		rc = gsm0808_dec_speech_codec_list(&conn->codec_list,
+						   TLVP_VAL(tp, GSM0808_IE_SPEECH_CODEC_LIST),
+						   TLVP_LEN(tp, GSM0808_IE_SPEECH_CODEC_LIST));
+		if (rc < 0) {
+			LOGP(DMSC, LOGL_ERROR, "Unable to decode speech codec list\n");
+			*cause = GSM0808_CAUSE_INCORRECT_VALUE;
+			return -1;
+		}
+	}
+
+	if (aoip && !conn->codec_list.len) {
+		LOGP(DMSC, LOGL_ERROR, "%s: AoIP Assignment Request:"
+		     " Missing or empty Speech Codec List IE\n", bsc_subscr_name(conn->bsub));
+		*cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
+		return -1;
+	}
+
+	return 0;
+}
+
+static int bssmap_handle_ass_req_tp_rtp_ext(struct tlv_parsed *tp,
+					    struct assignment_request *req,
+					    uint8_t *cause)
+{
+	/* If the special TW-TS-003 BSSMAP IE is not included,
+	 * we are in standard 3GPP-compliant operation mode -
+	 * no more work is needed here. */
+	if (!TLVP_PRESENT(tp, GSM0808_IE_THEMWI_RTP_EXTENSIONS))
+		return 0;
+
+	/* the format is TLV with one required byte of value */
+	if (TLVP_LEN(tp, GSM0808_IE_THEMWI_RTP_EXTENSIONS) < 1) {
+		*cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
+		return -1;
+	}
+
+	req->rtp_extensions = *TLVP_VAL(tp, GSM0808_IE_THEMWI_RTP_EXTENSIONS);
+	return 0;
+}
+
+static int bssmap_handle_ass_req_ct_data(struct gsm_subscriber_connection *conn, struct tlv_parsed *tp,
+					 struct gsm0808_channel_type *ct, struct assignment_request *req,
+					 uint8_t *cause)
+{
+	bool aoip = gscon_is_aoip(conn);
+	int rc;
+
+	*req = (struct assignment_request){
+		.assign_for = ASSIGN_FOR_BSSMAP_REQ,
+		.aoip = aoip,
+	};
+
+	if (bssmap_handle_ass_req_tp_cic(tp, aoip, &req->msc_assigned_cic, cause) < 0)
+		return -1;
+
+	if (bssmap_handle_ass_req_tp_rtp_addr(tp, aoip, req->msc_rtp_addr, sizeof(req->msc_rtp_addr), &req->msc_rtp_port, cause) < 0)
+		return -1;
+
+	/* According to 3GPP TS 48.008 § 3.2.1.1 note 13, the codec list IE
+	 * shall be included for aoip unless channel type is signalling. */
+	if (bssmap_handle_ass_req_tp_codec_list(conn, tp, aoip, cause) < 0)
+		return -1;
+
+	rc = select_data_rates(req, ct, conn);
+	if (rc < 0) {
+		*cause = GSM0808_CAUSE_REQ_CODEC_TYPE_OR_CONFIG_UNAVAIL;
+		return -1;
+	}
+
+	return 0;
+}
+
+int bssmap_handle_ass_req_ct_speech(struct gsm_subscriber_connection *conn, struct gsm_bts *bts,
+				    struct tlv_parsed *tp, struct gsm0808_channel_type *ct,
+				    struct assignment_request *req, uint8_t *cause)
+{
+	bool aoip = gscon_is_aoip(conn);
+	int rc;
+
+	*req = (struct assignment_request){
+		.assign_for = ASSIGN_FOR_BSSMAP_REQ,
+		.aoip = aoip,
+	};
+
+	if (bssmap_handle_ass_req_tp_cic(tp, aoip, &req->msc_assigned_cic, cause) < 0)
+		return -1;
+
+	if (bssmap_handle_ass_req_tp_rtp_addr(tp, aoip, req->msc_rtp_addr, sizeof(req->msc_rtp_addr), &req->msc_rtp_port, cause) < 0)
+		return -1;
+
+	if (bssmap_handle_ass_req_tp_osmux(conn, tp, &req->use_osmux, &req->osmux_cid, cause) < 0)
+		return -1;
+
+	if (bssmap_handle_ass_req_tp_codec_list(conn, tp, aoip, cause) < 0)
+		return -1;
+
+	/* Match codec information from the assignment command against the
+	 * local preferences of the BSC and BTS */
+	rc = select_codecs(req, ct, conn, bts);
+	if (rc < 0) {
+		*cause = GSM0808_CAUSE_REQ_CODEC_TYPE_OR_CONFIG_UNAVAIL;
+		return -1;
+	}
+
+	/* optional RTP extensions per TW-TS-003 */
+	if (bssmap_handle_ass_req_tp_rtp_ext(tp, req, cause) < 0)
+		return -1;
+
+	return 0;
+}
+
+static int bssmap_handle_ass_req_ct_sign(struct gsm_subscriber_connection *conn, struct gsm0808_channel_type *ct,
+					 struct assignment_request *req, uint8_t *cause)
+{
+	int rc;
+
+	*req = (struct assignment_request){
+		.assign_for = ASSIGN_FOR_BSSMAP_REQ,
+		.aoip = gscon_is_aoip(conn),
+	};
+
+	rc = select_sign_chan(req, ct);
+	if (rc < 0) {
+		*cause = GSM0808_CAUSE_INCORRECT_VALUE;
+		return rc;
+	}
+
+	return 0;
+}
+
 /*
  * Handle the assignment request message.
  *
@@ -806,12 +1112,8 @@ static int bssmap_handle_assignm_req(struct gsm_subscriber_connection *conn,
 {
 	struct msgb *resp;
 	struct tlv_parsed tp;
-	uint16_t cic = 0;
-	bool aoip = false;
-	bool use_osmux = false;
-	uint8_t osmux_cid = 0;
-	struct sockaddr_storage rtp_addr;
 	struct gsm0808_channel_type ct;
+	struct gsm0808_group_callref gc;
 	uint8_t cause;
 	int rc;
 	struct assignment_request req = {};
@@ -822,8 +1124,6 @@ static int bssmap_handle_assignm_req(struct gsm_subscriber_connection *conn,
 		return -1;
 	}
 
-	aoip = gscon_is_aoip(conn);
-
 	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
 		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
 		return -1;
@@ -845,149 +1145,71 @@ static int bssmap_handle_assignm_req(struct gsm_subscriber_connection *conn,
 		goto reject;
 	}
 
-	bssmap_handle_ass_req_lcls(conn, &tp);
+	/* Check for assignment to VGCS channel. */
+	if (TLVP_PRESENT(&tp, GSM0808_IE_GROUP_CALL_REFERENCE)) {
+		struct gsm_bts *bts = conn_get_bts(conn);
 
-	/* Currently we only support a limited subset of all
-	 * possible channel types, such as multi-slot or CSD */
-	switch (ct.ch_indctr) {
-	case GSM0808_CHAN_DATA:
-		LOGP(DMSC, LOGL_ERROR, "Unsupported channel type, currently only speech is supported!\n");
-		cause = GSM0808_CAUSE_REQ_CODEC_TYPE_OR_CONFIG_NOT_SUPP;
-		goto reject;
-	case GSM0808_CHAN_SPEECH:
-		if (TLVP_PRESENT(&tp, GSM0808_IE_CIRCUIT_IDENTITY_CODE)) {
-			/* CIC is permitted in both AoIP and SCCPlite */
-			cic = osmo_load16be(TLVP_VAL(&tp, GSM0808_IE_CIRCUIT_IDENTITY_CODE));
-		} else {
-			if (!aoip) {
-				/* no CIC but SCCPlite: illegal */
-				LOGP(DMSC, LOGL_ERROR, "SCCPlite MSC, but no CIC in ASSIGN REQ?\n");
-				cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
-				goto reject;
-			}
-		}
-		if (TLVP_PRESENT(&tp, GSM0808_IE_AOIP_TRASP_ADDR)) {
-			if (!aoip) {
-				/* SCCPlite and AoIP transport address: illegal */
-				LOGP(DMSC, LOGL_ERROR, "AoIP Transport address over IPA ?!?\n");
-				cause = GSM0808_CAUSE_INCORRECT_VALUE;
-				goto reject;
-			}
-			/* Decode AoIP transport address element */
-			rc = gsm0808_dec_aoip_trasp_addr(&rtp_addr,
-							 TLVP_VAL(&tp, GSM0808_IE_AOIP_TRASP_ADDR),
-							 TLVP_LEN(&tp, GSM0808_IE_AOIP_TRASP_ADDR));
-			if (rc < 0) {
-				LOGP(DMSC, LOGL_ERROR, "Unable to decode AoIP transport address.\n");
-				cause = GSM0808_CAUSE_INCORRECT_VALUE;
-				goto reject;
-			}
-		} else if (aoip) {
-			/* no AoIP transport level address but AoIP transport: illegal */
-			LOGP(DMSC, LOGL_ERROR, "AoIP transport address missing in ASSIGN REQ, "
-			     "audio would not work; rejecting\n");
-			cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
-			goto reject;
-		}
-
-		if (TLVP_PRESENT(&tp, GSM0808_IE_OSMO_OSMUX_CID)) {
-			if (conn->sccp.msc->use_osmux == OSMUX_USAGE_OFF) {
-				LOGP(DMSC, LOGL_ERROR, "MSC using Osmux but we have it disabled.\n");
-				cause = GSM0808_CAUSE_INCORRECT_VALUE;
-				goto reject;
-			}
-			use_osmux = true;
-			rc = gsm0808_dec_osmux_cid(&osmux_cid,
-						   TLVP_VAL(&tp, GSM0808_IE_OSMO_OSMUX_CID),
-						   TLVP_LEN(&tp, GSM0808_IE_OSMO_OSMUX_CID));
-			if (rc < 0) {
-				LOGP(DMSC, LOGL_ERROR, "Unable to decode Osmux CID.\n");
-				cause = GSM0808_CAUSE_INCORRECT_VALUE;
-				goto reject;
-			}
-		} else {
-			if (conn->sccp.msc->use_osmux == OSMUX_USAGE_ONLY) {
-				LOGP(DMSC, LOGL_ERROR, "MSC not using Osmux but we are forced to use it.\n");
-				cause = GSM0808_CAUSE_INCORRECT_VALUE;
-				goto reject;
-			} else if (conn->sccp.msc->use_osmux == OSMUX_USAGE_ON)
-				LOGP(DMSC, LOGL_NOTICE, "MSC not using Osmux but we have Osmux enabled.\n");
-		}
-
-		/* Decode speech codec list. First set len = 0. */
-		conn->codec_list = (struct gsm0808_speech_codec_list){};
-		/* Check for speech codec list element */
-		if (TLVP_PRESENT(&tp, GSM0808_IE_SPEECH_CODEC_LIST)) {
-			/* Decode Speech Codec list */
-			rc = gsm0808_dec_speech_codec_list(&conn->codec_list,
-							   TLVP_VAL(&tp, GSM0808_IE_SPEECH_CODEC_LIST),
-							   TLVP_LEN(&tp, GSM0808_IE_SPEECH_CODEC_LIST));
-			if (rc < 0) {
-				LOGP(DMSC, LOGL_ERROR, "Unable to decode speech codec list\n");
-				cause = GSM0808_CAUSE_INCORRECT_VALUE;
-				goto reject;
-			}
-		}
-
-		if (aoip && !conn->codec_list.len) {
-			LOGP(DMSC, LOGL_ERROR, "%s: AoIP speech mode Assignment Request:"
-			     " Missing or empty Speech Codec List IE\n", bsc_subscr_name(conn->bsub));
-			cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
-			goto reject;
-		}
-
-		req = (struct assignment_request){
-			.assign_for = ASSIGN_FOR_BSSMAP_REQ,
-			.aoip = aoip,
-			.msc_assigned_cic = cic,
-			.use_osmux = use_osmux,
-			.osmux_cid = osmux_cid,
-		};
-
-		/* Match codec information from the assignment command against the
-		 * local preferences of the BSC and BTS */
-		rc = select_codecs(&req, &ct, conn);
-		if (rc < 0) {
-			cause = GSM0808_CAUSE_REQ_CODEC_TYPE_OR_CONFIG_UNAVAIL;
-			goto reject;
-		}
-
-		if (aoip) {
-			unsigned int rc = osmo_sockaddr_to_str_and_uint(req.msc_rtp_addr,
-									sizeof(req.msc_rtp_addr),
-									&req.msc_rtp_port,
-									(const struct sockaddr*)&rtp_addr);
-			if (!rc || rc >= sizeof(req.msc_rtp_addr)) {
-				LOGP(DMSC, LOGL_ERROR, "Assignment request: RTP address is too long\n");
-				cause = GSM0808_CAUSE_REQ_CODEC_TYPE_OR_CONFIG_UNAVAIL;
-				goto reject;
-			}
-		}
-		break;
-	case GSM0808_CHAN_SIGN:
-		req = (struct assignment_request){
-			.assign_for = ASSIGN_FOR_BSSMAP_REQ,
-			.aoip = aoip,
-		};
-
-		rc = select_sign_chan(&req, &ct);
+		OSMO_ASSERT(bts);
+		/* Decode Group Call Reference. */
+		rc = gsm0808_dec_group_callref(&gc, TLVP_VAL(&tp, GSM0808_IE_GROUP_CALL_REFERENCE),
+					       TLVP_LEN(&tp, GSM0808_IE_GROUP_CALL_REFERENCE));
 		if (rc < 0) {
+			LOGP(DMSC, LOGL_ERROR, "Unable to decode Group Call Reference.\n");
 			cause = GSM0808_CAUSE_INCORRECT_VALUE;
 			goto reject;
 		}
+		req.target_lchan = vgcs_vbs_find_lchan(bts, &gc);
+		if (!req.target_lchan) {
+			cause = GSM0808_CAUSE_INCORRECT_VALUE;
+			goto reject;
+		}
+		req.assign_for = ASSIGN_FOR_BSSMAP_REQ;
+		req.vgcs = true;
+
+		/* Copy timing advance. */
+		if (conn->lchan) {
+			req.target_lchan->activate.info.ta_known = conn->lchan->activate.info.ta_known;
+			req.target_lchan->activate.info.ta = conn->lchan->activate.info.ta;
+		}
+
+		/* Send reactivation on target lchan to prepare VGCS channel for assignment.
+		 * See patent EP 1 858 275 A1. */
+		rsl_tx_chan_activ(req.target_lchan, RSL_ACT_TYPE_REACT | RSL_ACT_INTRA_NORM_ASS, 0);
+
+		return osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_ASSIGNMENT_START, &req);
+	}
+
+	bssmap_handle_ass_req_lcls(conn, &tp);
+
+	/* Currently we only support a limited subset of all
+	 * possible channel types, such as multi-slot */
+	switch (ct.ch_indctr) {
+	case GSM0808_CHAN_DATA:
+		if (bssmap_handle_ass_req_ct_data(conn, &tp, &ct, &req, &cause) < 0)
+			goto reject;
+		break;
+	case GSM0808_CHAN_SPEECH:
+		if (bssmap_handle_ass_req_ct_speech(conn, conn_get_bts(conn), &tp, &ct, &req, &cause) < 0)
+			goto reject;
+		break;
+	case GSM0808_CHAN_SIGN:
+		if (bssmap_handle_ass_req_ct_sign(conn, &ct, &req, &cause) < 0)
+			goto reject;
 		break;
 	default:
 		cause = GSM0808_CAUSE_INVALID_MESSAGE_CONTENTS;
 		goto reject;
 	}
 
+	req.ch_indctr = ct.ch_indctr;
+
 	return osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_ASSIGNMENT_START, &req);
 
 reject:
 	resp = gsm0808_create_assignment_failure(cause, NULL);
 	OSMO_ASSERT(resp);
 
-	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_ASSIGMENT_FAILURE));
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_ASSIGNMENT_FAILURE));
 	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
 	return -1;
 }
@@ -1135,6 +1357,101 @@ static int bssmap_handle_common_id(struct gsm_subscriber_connection *conn,
 	return 0;
 }
 
+/* Handle (VGCS) UPLINK REQUEST ACKNOWLEDGE:
+ *
+ * See 3GPP TS 48.008 §3.2.1.58
+ */
+static int bssmap_handle_uplink_rqst_acknowledge(struct gsm_subscriber_connection *conn,
+						 struct msgb *msg, unsigned int length)
+{
+	struct tlv_parsed tp;
+
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
+
+	if (conn->vgcs_call.fi)
+		osmo_fsm_inst_dispatch(conn->vgcs_call.fi, VGCS_EV_MSC_ACK, NULL);
+	return 0;
+}
+
+/* Handle (VGCS) UPLINK REJECT COMMAND message.
+ *
+ * See 3GPP TS 48.008 §3.2.1.61
+ */
+static int bssmap_handle_uplink_reject_cmd(struct gsm_subscriber_connection *conn,
+					   struct msgb *msg, unsigned int length)
+{
+	struct tlv_parsed tp;
+
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
+
+	if (conn->vgcs_call.fi)
+		osmo_fsm_inst_dispatch(conn->vgcs_call.fi, VGCS_EV_MSC_REJECT, NULL);
+	return 0;
+}
+
+/* Handle (VGCS) UPLINK RELEASE COMMAND message, MSC indicating an error to us:
+ *
+ * See 3GPP TS 48.008 §3.2.1.62
+ */
+static int bssmap_handle_uplink_release_cmd(struct gsm_subscriber_connection *conn,
+					    struct msgb *msg, unsigned int length)
+{
+	struct tlv_parsed tp;
+
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
+
+	if (conn->vgcs_call.fi)
+		osmo_fsm_inst_dispatch(conn->vgcs_call.fi, VGCS_EV_MSC_RELEASE, NULL);
+	return 0;
+}
+
+/* Handle (VGCS) UPLINK SEIZED COMMAND message:
+ *
+ * See 3GPP TS 48.008 §3.2.1.63
+ */
+static int bssmap_handle_uplink_seized_cmd(struct gsm_subscriber_connection *conn,
+					   struct msgb *msg, unsigned int length)
+{
+	struct tlv_parsed tp;
+
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
+
+	if (conn->vgcs_call.fi)
+		osmo_fsm_inst_dispatch(conn->vgcs_call.fi, VGCS_EV_MSC_SEIZE, NULL);
+	return 0;
+}
+
+/* Handle VGCS/VBS ADDITIONAL INFO message:
+ *
+ * See 3GPP TS 48.008 §3.2.1.78
+ */
+static int bssmap_handle_vgcs_addl_info(struct gsm_subscriber_connection *conn,
+					struct msgb *msg, unsigned int length)
+{
+	struct tlv_parsed tp;
+
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
+
+	LOGPFSML(conn->fi, LOGL_ERROR, "VGCS ADDITIONAL INFO is not supported.\n");
+
+	return 0;
+}
+
 static int bssmap_rcvmsg_udt(struct bsc_msc_data *msc,
 			     struct msgb *msg, unsigned int length)
 {
@@ -1238,6 +1555,26 @@ static int bssmap_rcvmsg_dt1(struct gsm_subscriber_connection *conn,
 			ret = 0;
 		}
 		break;
+	case BSS_MAP_MSG_UPLINK_RQST_ACKNOWLEDGE:
+		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_UPLINK_RQST_ACKNOWLEDGE]);
+		ret = bssmap_handle_uplink_rqst_acknowledge(conn, msg, length);
+		break;
+	case BSS_MAP_MSG_UPLINK_REJECT_CMD:
+		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_UPLINK_REJECT_CMD]);
+		ret = bssmap_handle_uplink_reject_cmd(conn, msg, length);
+		break;
+	case BSS_MAP_MSG_UPLINK_RELEASE_CMD:
+		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_UPLINK_RELEASE_CMD]);
+		ret = bssmap_handle_uplink_release_cmd(conn, msg, length);
+		break;
+	case BSS_MAP_MSG_UPLINK_SEIZED_CMD:
+		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_UPLINK_SEIZED_CMD]);
+		ret = bssmap_handle_uplink_seized_cmd(conn, msg, length);
+		break;
+	case BSS_MAP_MSG_VGCS_ADDL_INFO:
+		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_VGCS_ADDL_INFO]);
+		ret = bssmap_handle_vgcs_addl_info(conn, msg, length);
+		break;
 	default:
 		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_UNKNOWN]);
 		LOGP(DMSC, LOGL_NOTICE, "Unimplemented msg type: %s\n",
@@ -1301,7 +1638,10 @@ static int dtap_rcvmsg(struct gsm_subscriber_connection *conn,
 	/* convert DLCI to RSL link ID, store in msg->cb */
 	OBSC_LINKID_CB(gsm48) = DLCI2RSL_LINK_ID(header->link_id);
 
-	dtap_rc = osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_MT_DTAP, gsm48);
+	if (conn->vgcs_call.fi)
+		dtap_rc = osmo_fsm_inst_dispatch(conn->vgcs_call.fi, VGCS_EV_MSC_DTAP, gsm48);
+	else
+		dtap_rc = osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_MT_DTAP, gsm48);
 	return dtap_rc;
 }
 
@@ -1461,7 +1801,7 @@ int bsc_tx_bssmap_ho_request_ack(struct gsm_subscriber_connection *conn, struct
 								  new_lchan->current_ch_mode_rate.chan_mode),
 	};
 
-	if (gscon_is_aoip(conn) && new_lchan->activate.info.requires_voice_stream) {
+	if (gscon_is_aoip(conn) && bsc_chan_ind_requires_rtp_stream(new_lchan->activate.info.ch_indctr)) {
 		struct osmo_sockaddr_str to_msc_rtp;
 		const struct mgcp_conn_peer *rtp_info = osmo_mgcpc_ep_ci_get_rtp_info(conn->user_plane.mgw_endpoint_ci_msc);
 		int rc;
@@ -1536,7 +1876,7 @@ enum handover_result bsc_tx_bssmap_ho_complete(struct gsm_subscriber_connection
 	};
 
 	/* speech_codec_chosen */
-	if (ho->new_lchan->activate.info.requires_voice_stream && gscon_is_aoip(conn)) {
+	if (bsc_chan_ind_requires_rtp_stream(ho->new_lchan->activate.info.ch_indctr) && gscon_is_aoip(conn)) {
 		int perm_spch = gsm0808_permitted_speech(lchan->type, lchan->current_ch_mode_rate.chan_mode);
 		params.speech_codec_chosen_present = true;
 		rc = gsm0808_speech_codec_from_chan_type(&params.speech_codec_chosen, perm_spch);
@@ -1580,3 +1920,180 @@ void bsc_tx_bssmap_ho_failure(struct gsm_subscriber_connection *conn)
 		LOG_HO(conn, LOGL_ERROR, "Cannot send BSSMAP Handover Failure message (rc=%d %s)\n",
 		       rc, strerror(-rc));
 }
+
+/* Send SETUP ACKNOWLEDGE to MSC. */
+void bsc_tx_setup_ack(struct gsm_subscriber_connection *conn, struct gsm0808_vgcs_feature_flags *ff)
+{
+	struct msgb *resp;
+	struct gsm0808_vgcs_vbs_setup_ack sa = {};
+
+	if (ff) {
+		sa.vgcs_feature_flags_present = true;
+		sa.flags = *ff;
+	}
+	resp = gsm0808_create_vgcs_vbs_setup_ack(&sa);
+	OSMO_ASSERT(resp);
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_VGCS_VBS_SETUP_ACK));
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
+}
+
+/* Send SETUP REFUSE to MSC. */
+void bsc_tx_setup_refuse(struct gsm_subscriber_connection *conn, uint8_t cause)
+{
+	struct msgb *resp;
+	resp = gsm0808_create_vgcs_vbs_setup_refuse(cause);
+
+	OSMO_ASSERT(resp);
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_VGCS_VBS_SETUP_REFUSE));
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
+}
+
+/* Send ASSIGNMENT FAILURE to MSC. */
+void bsc_tx_vgcs_vbs_assignment_fail(struct gsm_subscriber_connection *conn, uint8_t cause)
+{
+	struct msgb *resp;
+	struct gsm0808_vgcs_vbs_assign_fail af = {
+		.cause = cause,
+	};
+
+	resp = gsm0808_create_vgcs_vbs_assign_fail(&af);
+	OSMO_ASSERT(resp);
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_VGCS_VBS_ASSIGN_FAIL));
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
+}
+
+/* Send ASSIGNMENT RESULT to MSC. */
+void bsc_tx_vgcs_vbs_assignment_result(struct gsm_subscriber_connection *conn, struct gsm0808_channel_type *ct,
+				       struct gsm0808_cell_id *ci, uint32_t call_id)
+{
+	struct gsm_lchan *lchan = conn->lchan;
+	struct msgb *resp;
+	struct gsm0808_vgcs_vbs_assign_res ar = {
+		.channel_type = *ct,
+		.cell_identifier = *ci,
+	};
+	int perm_spch;
+	uint8_t osmux_cid;
+
+	/* Chosen Channel */
+	ar.chosen_channel = gsm0808_chosen_channel(lchan->type, lchan->current_ch_mode_rate.chan_mode);
+	if (!ar.chosen_channel) {
+		LOGP(DMSC, LOGL_ERROR, "Unable to compose Chosen Channel for mode=%s type=%s",
+		     get_value_string(gsm48_chan_mode_names, lchan->current_ch_mode_rate.chan_mode),
+		     gsm_chan_t_name(lchan->type));
+		bsc_tx_vgcs_vbs_assignment_fail(conn, GSM0808_CAUSE_EQUIPMENT_FAILURE);
+		return;
+	}
+	ar.chosen_channel_present = true;
+
+	/* Generate RTP related fields. */
+	if (gscon_is_aoip(conn)) {
+		/* AoIP Transport Layer Address (BSS) */
+		if (!osmo_mgcpc_ep_ci_get_crcx_info_to_sockaddr(conn->user_plane.mgw_endpoint_ci_msc,
+								&ar.aoip_transport_layer)) {
+			LOGP(DMSC, LOGL_ERROR, "Unable to compose RTP address of MGW -> MSC");
+			bsc_tx_vgcs_vbs_assignment_fail(conn, GSM0808_CAUSE_EQUIPMENT_FAILURE);
+			return;
+		}
+		ar.aoip_transport_layer_present = true;
+
+		/* Call Identifier */
+		ar.call_id = call_id;
+		ar.call_id_present = true;
+
+		/* Osmux */
+		if (conn->assignment.req.use_osmux) {
+			if (!osmo_mgcpc_ep_ci_get_crcx_info_to_osmux_cid(conn->user_plane.mgw_endpoint_ci_msc,
+									 &osmux_cid)) {
+				LOGP(DMSC, LOGL_ERROR, "Unable to compose Osmux CID of MGW -> MSC");
+				bsc_tx_vgcs_vbs_assignment_fail(conn, GSM0808_CAUSE_EQUIPMENT_FAILURE);
+				return;
+			}
+		}
+
+		/* Extrapolate speech codec from speech mode */
+		perm_spch = gsm0808_permitted_speech(lchan->type, lchan->current_ch_mode_rate.chan_mode);
+		gsm0808_speech_codec_from_chan_type(&ar.codec_msc_chosen, perm_spch);
+		ar.codec_msc_chosen.cfg = conn->lchan->current_ch_mode_rate.s15_s0;
+		ar.codec_present = true;
+	}
+
+	resp = gsm0808_create_vgcs_vbs_assign_res(&ar);
+	OSMO_ASSERT(resp);
+	if (conn->assignment.req.use_osmux)
+		bssap_extend_osmux(resp, osmux_cid);
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_VGCS_VBS_ASSIGN_RESULT));
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
+}
+
+/* Send UPLINK REQUEST to MSC. */
+void bsc_tx_uplink_req(struct gsm_subscriber_connection *conn)
+{
+	struct msgb *resp;
+	struct gsm0808_uplink_request ur = {};
+
+	resp = gsm0808_create_uplink_request(&ur);
+	OSMO_ASSERT(resp);
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_UPLINK_RQST));
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
+}
+
+/* Send UPLINK REQUEST CONFIRMATION to MSC. */
+void bsc_tx_uplink_req_conf(struct gsm_subscriber_connection *conn, struct gsm0808_cell_id *ci, uint8_t *l3_info,
+			    uint8_t length)
+{
+	struct msgb *resp;
+	struct gsm0808_uplink_request_cnf ur = {
+		.cell_identifier = *ci,
+	};
+
+	OSMO_ASSERT(length <= LAYER_3_INFORMATION_MAXLEN);
+	if (length) {
+		memcpy(ur.l3.l3, l3_info, length);
+		ur.l3.l3_len = length;
+	}
+	resp = gsm0808_create_uplink_request_cnf(&ur);
+	OSMO_ASSERT(resp);
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_UPLINK_RQST_CONFIRMATION));
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
+}
+
+/* Send UPLINK APPLICATION DATA to MSC. */
+void bsc_tx_uplink_app_data(struct gsm_subscriber_connection *conn, struct gsm0808_cell_id *ci, uint8_t *l3_info,
+			    uint8_t length)
+{
+	struct msgb *resp;
+	struct gsm0808_uplink_app_data ad = {
+		.cell_identifier = *ci,
+	};
+
+	OSMO_ASSERT(length <= LAYER_3_INFORMATION_MAXLEN);
+	memcpy(ad.l3.l3, l3_info, length);
+	ad.l3.l3_len = length;
+	resp = gsm0808_create_uplink_app_data(&ad);
+	OSMO_ASSERT(resp);
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_UPLINK_APP_DATA));
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
+}
+
+/* Send UPLINK RELEASE INDICATION to MSC. */
+void bsc_tx_uplink_release_ind(struct gsm_subscriber_connection *conn, uint8_t cause)
+{
+	struct msgb *resp;
+	struct gsm0808_uplink_release_ind ri = {
+		.cause = cause,
+	};
+
+	resp = gsm0808_create_uplink_release_ind(&ri);
+	OSMO_ASSERT(resp);
+
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_UPLINK_RELEASE_INDICATION));
+	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_TX_SCCP, resp);
+}
