@@ -376,26 +376,15 @@ static const struct osmo_stat_item_group_desc msc_statg_desc = {
 
 int osmo_bsc_msc_init(struct bsc_msc_data *msc)
 {
-	struct gsm_network *net = msc->network;
-	struct mgcp_client *mgcp_cli;
 	int rc;
 
 	/* Everything below refers to SCCP-Lite MSC connections only. */
 	if (msc_is_aoip(msc))
 		return 0;
 
-	/* Note: MGW is preselected here at startup, which means currently
-	 * osmo-bsc configured for SCCPLite doesn't support MGW pools with more
-	 * than 1 MGW.
-	 */
-	mgcp_cli = mgcp_client_pool_get(net->mgw.mgw_pool);
-	OSMO_ASSERT(mgcp_cli);
 	rc = osmo_sock_init2_ofd(&msc->mgcp_ipa.ofd, AF_UNSPEC, SOCK_DGRAM, IPPROTO_UDP,
 				 msc->mgcp_ipa.local_addr, msc->mgcp_ipa.local_port,
-				 mgcp_client_remote_addr_str(mgcp_cli),
-				 mgcp_client_remote_port(mgcp_cli),
-				 OSMO_SOCK_F_BIND | OSMO_SOCK_F_CONNECT);
-	mgcp_client_pool_put(mgcp_cli);
+				 NULL, 0, OSMO_SOCK_F_BIND);
 	if (rc < 0) {
 		LOGP(DMSC, LOGL_ERROR, "msc %u: Could not create/connect/bind MGCP proxy socket: %d\n",
 			msc->nr, rc);
@@ -454,11 +443,18 @@ struct bsc_msc_data *osmo_msc_data_alloc(struct gsm_network *net, int nr)
 	};
 
 	msc_data->nr = nr;
-	msc_data->allow_emerg = 1;
+	msc_data->allow_emerg = 0;
 	msc_data->a.asp_proto = OSMO_SS7_ASP_PROT_M3UA;
 
 	/* Defaults for the audio setup */
+	msc_data->amr_conf.m12_2 = 1;
+	msc_data->amr_conf.m10_2 = 1;
+	msc_data->amr_conf.m7_95 = 1;
+	msc_data->amr_conf.m7_40 = 1;
+	msc_data->amr_conf.m6_70 = 1;
 	msc_data->amr_conf.m5_90 = 1;
+	msc_data->amr_conf.m5_15 = 1;
+	msc_data->amr_conf.m4_75 = 1;
 	msc_data->amr_octet_aligned = true;
 
 	/* Allow the full set of possible codecs by default */
